%{ open Ast 
   open Lexing

let parse_error msg  = 
  raise ( Exceptions.ParserException( ((rhs_start_pos 1).pos_lnum, ((rhs_start_pos 1).pos_cnum -(rhs_start_pos 1).pos_bol), (rhs_end_pos 1).pos_lnum, ((rhs_end_pos 1).pos_cnum - (rhs_end_pos 1).pos_bol), msg)))
%}

 
%token SEMI LPAREN RPAREN LBRACE RBRACE COMMA
%token PLUS MINUS TIMES DIVIDE ASSIGN NOT LET
%token LBRACKET RBRACKET UNDER CONCAT
%token EQ NEQ LT LEQ GT GEQ AND OR
%token RETURN IF ELSE FOR WHILE
%token FLOAT STRING VOID INT

%token <string * int> ILITERAL
%token <string * float> FLITERAL
%token <string * string> SLITERAL
%token <string * string> ID
%token <string * string> COMMENT
%token EOF


%nonassoc NOELSE
%nonassoc ELSE
%right ASSIGN
%left OR
%left AND
%left EQ NEQ
%left LT GT LEQ GEQ
%left PLUS MINUS CONCAT
%left TIMES DIVIDE
%right NOT NEG

 
%start program
%type <Ast.program> program
 
%%
 
program: decls EOF { ((List.rev (fst $1)), (List.rev (snd $1))) }
 
decls: /* nothing */ { [], [] }
  | decls vdecl SEMI      { ($2 :: fst $1), snd $1 }
  | decls fdecl      { fst $1, ($2 :: snd $1) }
 
fdecl: comment typ ID LPAREN formals_opt RPAREN comment LBRACE stmt_list RBRACE
 	{ { comm = List.rev($1); typ = $2; fname = snd $3; formals = List.rev($5);
     	    body = Block($7, List.rev($9)) } }

comment: /* nothing */ { [] }
  | comment COMMENT { $2 :: $1 }
 
formals_opt: /* nothing */ { [] }
  | var_typ ID                 { [($1,snd $2)] }
  | formals_opt COMMA var_typ ID  { ($3,snd $4) :: $1 }

typ: FLOAT { Float }
   | STRING {String}
   | VOID { Void }
   | INT { Int }
   | ID UNDER LBRACE qual_list_expr { TensorType($4) }

var_typ: FLOAT { Float }
  | STRING { String }
  | INT { Int }
  | ID UNDER LBRACE qual_list_expr { TensorType($4) } 

vdecl: LET ID UNDER LBRACE qual_list_expr { TensorBind(fst $2, snd $2, $5)}
   | LET ID UNDER LBRACE qual_list_expr ASSIGN expr { TensorBindAndAssign(fst $2, snd $2, $5, $7) }
   | LET ID ASSIGN expr { InferredBind(fst $2, snd $2, $4) }

stmt_list: /* nothing */ {[] }
  | stmt_list stmt { $2 :: $1 }
 
stmt: comment expr SEMI							{ Expr (List.rev($1), $2) }
  | comment RETURN SEMI							{ Return (List.rev($1), Noexpr) }
  | comment RETURN expr SEMI                    	  	 	{ Return(List.rev($1), $3) }
  | comment LBRACE stmt_list RBRACE				{ Block(List.rev($1), List.rev($3)) }
  | comment IF LPAREN expr RPAREN stmt %prec NOELSE 			{ If(List.rev($1), $4, $6, Block([], [])) }
  | comment IF LPAREN expr RPAREN stmt ELSE stmt	     	 	{ If(List.rev($1), $4, $6, $8) }
  | comment FOR LPAREN expr_opt SEMI expr SEMI expr_opt RPAREN stmt  	{ For(List.rev($1), $4, $6, $8, $10) }
  | comment WHILE LPAREN expr RPAREN stmt       			{ While(List.rev($1), $4, $6) }
	 

qual_list_expr: expr RBRACE		{ [$1] } 
  | expr COMMA qual_list_expr 		{ $1::$3 }
 

ten_list: expr COMMA ten_list { $1::$3 }
  | expr       	     	      { [$1] }


expr:
    ILITERAL			 { Iliteral(fst $1, snd $1) }
  | SLITERAL                  	 { Sliteral(fst $1, snd $1) }
  | FLITERAL			 { Fliteral(fst $1, snd $1) }
  | expr PLUS expr       	 { Binop(get_line_expr $1, $1, Add, $3) }
  | expr MINUS expr         	 { Binop(get_line_expr $1, $1, Sub, $3) }
  | expr TIMES expr         	 { Binop(get_line_expr $1, $1, Mult, $3) }
  | expr DIVIDE expr         	 { Binop(get_line_expr $1, $1, Div, $3) }
  | expr CONCAT expr		 { Binop(get_line_expr $1, $1, Concat, $3) }
  | expr EQ expr     	        	 { Binop(get_line_expr $1, $1, Equal, $3) }
  | expr NEQ expr  	        	 { Binop(get_line_expr $1, $1, Neq, $3) }
  | expr LT expr     	        	 { Binop(get_line_expr $1, $1, Less, $3) }
  | expr LEQ expr   	        	 { Binop(get_line_expr $1, $1, Leq, $3) }
  | expr GT expr     	        	 { Binop(get_line_expr $1, $1, Greater, $3) }
  | expr GEQ expr  	        	 { Binop(get_line_expr $1, $1, Geq, $3) }
  | expr AND expr  	        	 { Binop(get_line_expr $1, $1, And, $3) }
  | expr OR expr     	        	 { Binop(get_line_expr $1, $1, Or, $3) }
  | MINUS expr %prec NEG 	 { Unop(get_line_expr $2, Neg, $2) }
  | PLUS expr			 { $2 (* Unary plus is redundant; ignore it *) }
  | NOT expr           	        	 { Unop(get_line_expr $2, Not, $2) }
  | ID					{ Id(fst $1, snd $1) }
  | vdecl	{ BindExpr(get_line_bind $1, $1) }
  | ID ASSIGN expr 	        	 { Assign(fst $1, snd $1, $3) }
  | LPAREN expr RPAREN 	 { $2 }
  | ID LPAREN actuals_opt RPAREN { Call(fst $1, snd $1, $3) }
  | ID UNDER LBRACE qual_list_expr { TensorId(fst $1, snd $1, $4) }
  | ID UNDER LBRACE qual_list_expr ASSIGN expr { TensorAssign(fst $1, snd $1, $4, $6) }
  | LBRACKET ten_list RBRACKET { Tensor(get_line_expr (List.hd $2), $2) }


expr_opt:
	/* nothing */ { Noexpr }
  | expr { $1 }
 
actuals_opt:
	/* nothing */ { [] }
  | actuals_list { List.rev $1 }
 
actuals_list:
	expr { [$1] }
| actuals_list COMMA expr { $3 :: $1 }
