{	open Parser
	open Lexing
	let filename = ref ""
	let lineno lb = 
		lb.lex_curr_p.pos_lnum

  let unescape s =
    Scanf.unescaped s
 }

let alpha = ['a'-'z' 'A'-'Z']
let ascii = ([' '-'!' '#'-'[' ']'-'~'])
let digit = ['0'-'9']
let id = alpha (alpha | digit)*
(*Fix this junk for floating points.*)
let float = digit* ('.' digit*)? (['e' 'E'] ['-' '+']? digit+)?
let int = digit+
let whitespace = [' ' '\t' '\r']
let escape = '\\' ['\\' '\'' '"' 'n' 'r' 't']
let string = '"' ((escape | [^ '"' '\\'])* as str) '"'



rule token = parse
  [' ' '\t' '\r'] { token lexbuf } (* Whitespace *)
| '\n' 		{ Lexing.new_line lexbuf; token lexbuf } (* Newline *)
| "%{"		{ comment lexbuf } (* Block comments *)
| "%"		{ singlecomment lexbuf } (* Single-line comments *)
| ("%%" [^ '\n' ]*) as lxm { COMMENT(string_of_int lexbuf.lex_curr_p.pos_lnum, match String.length lxm with 0 | 1 | 2 -> "" | x -> String.sub lxm 2 (x-2)) }
| '('		{ LPAREN }
| ')'		{ RPAREN }
| '['		{ LBRACKET }
| ']' 		{ RBRACKET }
| '{'		{ LBRACE }
| '}'		{ RBRACE }
| ';'		{ SEMI }
| ','		{ COMMA }
| '_'		{ UNDER }
| '+'		{ PLUS }
| '-'		{ MINUS }
| '*'		{ TIMES }
| '/'		{ DIVIDE }
| '^'		{ CONCAT }
| "=="		{ EQ }
| '='		{ ASSIGN }
| "~="		{ NEQ }
| '<'		{ LT }
| "<="		{ LEQ }
| '>'		{ GT }
| ">="		{ GEQ }
| "&&"		{ AND }
| "||"		{ OR }
| '~'		{ NOT }
| "else"	{ ELSE }
| "int"		{ INT }
| "float"	{ FLOAT }
| "void"  { VOID }
| "for"		{ FOR }
| "if"		{ IF }
| "let"		{ LET }
| "return"	{ RETURN }
| "string"	{ STRING }
| "while"	{ WHILE }
| int as lxm { ILITERAL(string_of_int (lineno lexbuf), int_of_string lxm) }
| float as lxm { FLITERAL(string_of_int (lineno lexbuf), float_of_string lxm) }
| string { 
	String.iter (fun n-> match n with '\n' -> Lexing.new_line lexbuf | _ -> ()) str;
	SLITERAL(string_of_int (lineno lexbuf), unescape str) }
| id as lxm { ID(string_of_int (lineno lexbuf), lxm) }
| eof { EOF }
| '"'			{ raise (Exceptions.UnmatchedQuotationException(lineno lexbuf)) }
| _ as illegal  { raise (Exceptions.IllegalCharacterException(lineno lexbuf, illegal)) }

and comment = parse
  "%}" { token lexbuf }
| '\n' { Lexing.new_line lexbuf; comment lexbuf } (* Increment line counter *)
| _ { comment lexbuf }

and singlecomment = parse
  ('\n') { Lexing.new_line lexbuf; token lexbuf } (* Single-line comment ends with a newline or a carriage return*)
| "%{" { comment lexbuf } (* If we see a multiline comment,
                             forget about looking for a newline *)
| _ { singlecomment lexbuf }
